@@ -1,5 +1,6 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
- * Copyright (C) 2005-2016 Junjiro R. Okajima
+ * Copyright (C) 2005-2018 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,19 +23,13 @@
 #include <linux/namei.h>
 #include "aufs.h"
 
-struct au_do_lookup_args {
-	unsigned int		flags;
-	mode_t			type;
-};
-
 /*
  * returns positive/negative dentry, NULL or an error.
  * NULL means whiteout-ed or not-found.
  */
 static struct dentry*
 au_do_lookup(struct dentry *h_parent, struct dentry *dentry,
-	     aufs_bindex_t bindex, struct qstr *wh_name,
-	     struct au_do_lookup_args *args)
+	     aufs_bindex_t bindex, struct au_do_lookup_args *args)
 {
 	struct dentry *h_dentry;
 	struct inode *h_inode;
@@ -49,7 +44,7 @@ au_do_lookup(struct dentry *h_parent, struct dentry *dentry,
 	br = au_sbr(dentry->d_sb, bindex);
 	wh_able = !!au_br_whable(br->br_perm);
 	if (wh_able)
-		wh_found = au_wh_test(h_parent, wh_name, ignore_perm);
+		wh_found = au_wh_test(h_parent, &args->whname, ignore_perm);
 	h_dentry = ERR_PTR(wh_found);
 	if (!wh_found)
 		goto real_lookup;
@@ -64,9 +59,9 @@ au_do_lookup(struct dentry *h_parent, struct dentry *dentry,
 
 real_lookup:
 	if (!ignore_perm)
-		h_dentry = vfsub_lkup_one(&dentry->d_name, h_parent);
+		h_dentry = vfsub_lkup_one(args->name, h_parent);
 	else
-		h_dentry = au_sio_lkup_one(&dentry->d_name, h_parent);
+		h_dentry = au_sio_lkup_one(args->name, h_parent);
 	if (IS_ERR(h_dentry)) {
 		if (PTR_ERR(h_dentry) == -ENAMETOOLONG
 		    && !allow_neg)
@@ -81,6 +76,13 @@ real_lookup:
 	} else if (wh_found
 		   || (args->type && args->type != (h_inode->i_mode & S_IFMT)))
 		goto out_neg;
+	else if (au_ftest_lkup(args->flags, DIRREN)
+		 /* && h_inode */
+		 && !au_dr_lkup_h_ino(args, bindex, h_inode->i_ino)) {
+		AuDbg("b%d %pd ignored hi%llu\n", bindex, h_dentry,
+		      (unsigned long long)h_inode->i_ino);
+		goto out_neg;
+	}
 
 	if (au_dbbot(dentry) <= bindex)
 		au_set_dbbot(dentry, bindex);
@@ -93,9 +95,9 @@ real_lookup:
 	    || (d_really_is_positive(dentry) && !d_is_dir(dentry)))
 		goto out; /* success */
 
-	inode_lock_nested(h_inode, AuLsc_I_CHILD);
+	inode_lock_shared_nested(h_inode, AuLsc_I_CHILD);
 	opq = au_diropq_test(h_dentry);
-	inode_unlock(h_inode);
+	inode_unlock_shared(h_inode);
 	if (opq > 0)
 		au_set_dbdiropq(dentry, bindex);
 	else if (unlikely(opq < 0)) {
@@ -129,26 +131,28 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t btop,
 {
 	int npositive, err;
 	aufs_bindex_t bindex, btail, bdiropq;
-	unsigned char isdir, dirperm1;
-	struct qstr whname;
+	unsigned char isdir, dirperm1, dirren;
 	struct au_do_lookup_args args = {
-		.flags		= flags
+		.flags		= flags,
+		.name		= &dentry->d_name
 	};
-	const struct qstr *name = &dentry->d_name;
 	struct dentry *parent;
 	struct super_block *sb;
 
 	sb = dentry->d_sb;
-	err = au_test_shwh(sb, name);
+	err = au_test_shwh(sb, args.name);
 	if (unlikely(err))
 		goto out;
 
-	err = au_wh_name_alloc(&whname, name);
+	err = au_wh_name_alloc(&args.whname, args.name);
 	if (unlikely(err))
 		goto out;
 
 	isdir = !!d_is_dir(dentry);
 	dirperm1 = !!au_opt_test(au_mntflags(sb), DIRPERM1);
+	dirren = !!au_opt_test(au_mntflags(sb), DIRREN);
+	if (dirren)
+		au_fset_lkup(args.flags, DIRREN);
 
 	npositive = 0;
 	parent = dget_parent(dentry);
@@ -156,6 +160,7 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t btop,
 	for (bindex = btop; bindex <= btail; bindex++) {
 		struct dentry *h_parent, *h_dentry;
 		struct inode *h_inode, *h_dir;
+		struct au_branch *br;
 
 		h_dentry = au_h_dptr(dentry, bindex);
 		if (h_dentry) {
@@ -167,11 +172,17 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t btop,
 		if (!h_parent || !d_is_dir(h_parent))
 			continue;
 
+		if (dirren) {
+			/* if the inum matches, then use the prepared name */
+			err = au_dr_lkup_name(&args, bindex);
+			if (unlikely(err))
+				goto out_parent;
+		}
+
 		h_dir = d_inode(h_parent);
-		inode_lock_nested(h_dir, AuLsc_I_PARENT);
-		h_dentry = au_do_lookup(h_parent, dentry, bindex, &whname,
-					&args);
-		inode_unlock(h_dir);
+		inode_lock_shared_nested(h_dir, AuLsc_I_PARENT);
+		h_dentry = au_do_lookup(h_parent, dentry, bindex, &args);
+		inode_unlock_shared(h_dir);
 		err = PTR_ERR(h_dentry);
 		if (IS_ERR(h_dentry))
 			goto out_parent;
@@ -198,6 +209,15 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t btop,
 			if (bdiropq >= 0 && bdiropq <= bindex)
 				break;
 		}
+		br = au_sbr(sb, bindex);
+		if (dirren
+		    && au_dr_hino_test_add(&br->br_dirren, h_inode->i_ino,
+					   /*add_ent*/NULL)) {
+			/* prepare next name to lookup */
+			err = au_dr_lkup(&args, dentry, bindex);
+			if (unlikely(err))
+				goto out_parent;
+		}
 	}
 
 	if (npositive) {
@@ -214,7 +234,9 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t btop,
 
 out_parent:
 	dput(parent);
-	au_delayed_kfree(whname.name);
+	kfree(args.whname.name);
+	if (dirren)
+		au_dr_lkup_fin(&args);
 out:
 	return err;
 }
@@ -301,7 +323,7 @@ static void au_iattr_save(struct au_iattr *ia, struct inode *h_inode)
 	/* ia->i_nlink = h_inode->i_nlink; */
 	ia->i_uid = h_inode->i_uid;
 	ia->i_gid = h_inode->i_gid;
-	ia->i_version = h_inode->i_version;
+	ia->i_version = inode_query_iversion(h_inode);
 /*
 	ia->i_size = h_inode->i_size;
 	ia->i_blocks = h_inode->i_blocks;
@@ -315,7 +337,7 @@ static int au_iattr_test(struct au_iattr *ia, struct inode *h_inode)
 		/* || ia->i_nlink != h_inode->i_nlink */
 		|| !uid_eq(ia->i_uid, h_inode->i_uid)
 		|| !gid_eq(ia->i_gid, h_inode->i_gid)
-		|| ia->i_version != h_inode->i_version
+		|| !inode_eq_iversion(h_inode, ia->i_version)
 /*
 		|| ia->i_size != h_inode->i_size
 		|| ia->i_blocks != h_inode->i_blocks
@@ -823,7 +845,7 @@ out:
 
 /* todo: remove this */
 static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
-			  unsigned int flags, int do_udba)
+			  unsigned int flags, int do_udba, int dirren)
 {
 	int err;
 	umode_t mode, h_mode;
@@ -874,7 +896,7 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 			     && !is_root
 			     && ((!h_nfs
 				  && (unhashed != !!d_unhashed(h_dentry)
-				      || (!tmpfile
+				      || (!tmpfile && !dirren
 					  && !au_qstreq(name, h_name))
 					  ))
 				 || (h_nfs
@@ -1015,7 +1037,7 @@ static int aufs_d_revalidate(struct dentry *dentry, unsigned int flags)
 {
 	int valid, err;
 	unsigned int sigen;
-	unsigned char do_udba;
+	unsigned char do_udba, dirren;
 	struct super_block *sb;
 	struct inode *inode;
 
@@ -1088,7 +1110,8 @@ static int aufs_d_revalidate(struct dentry *dentry, unsigned int flags)
 		}
 	}
 
-	err = h_d_revalidate(dentry, inode, flags, do_udba);
+	dirren = !!au_opt_test(au_mntflags(sb), DIRREN);
+	err = h_d_revalidate(dentry, inode, flags, do_udba, dirren);
 	if (unlikely(!err && do_udba && au_dbtop(dentry) < 0)) {
 		err = -EIO;
 		AuDbg("both of real entry and whiteout found, %p, err %d\n",
