@@ -1,5 +1,6 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
- * Copyright (C) 2005-2016 Junjiro R. Okajima
+ * Copyright (C) 2005-2018 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -156,7 +157,7 @@ out_unlock:
 out:
 	dput(a->dentry);
 	au_nwt_done(&au_sbi(sb)->si_nowait);
-	au_delayed_kfree(arg);
+	kfree(arg);
 }
 
 void au_dir_ts(struct inode *dir, aufs_bindex_t bindex)
@@ -192,7 +193,7 @@ void au_dir_ts(struct inode *dir, aufs_bindex_t bindex)
 	if (unlikely(wkq_err)) {
 		pr_err("wkq %d\n", wkq_err);
 		dput(dentry);
-		au_delayed_kfree(arg);
+		kfree(arg);
 	}
 
 out:
@@ -258,7 +259,7 @@ static int do_open_dir(struct file *file, int flags, struct file *h_file)
 	err = 0;
 	mnt = file->f_path.mnt;
 	dentry = file->f_path.dentry;
-	file->f_version = d_inode(dentry)->i_version;
+	file->f_version = inode_query_iversion(d_inode(dentry));
 	bindex = au_dbtop(dentry);
 	au_set_fbtop(file, bindex);
 	btail = au_dbtaildir(dentry);
@@ -311,7 +312,7 @@ static int aufs_open_dir(struct inode *inode __maybe_unused,
 		};
 		err = au_do_open(file, &args);
 		if (unlikely(err))
-			au_delayed_kfree(fidir);
+			kfree(fidir);
 	}
 	si_read_unlock(sb);
 	return err;
@@ -325,21 +326,18 @@ static int aufs_release_dir(struct inode *inode __maybe_unused,
 	struct au_fidir *fidir;
 	struct au_hfile *hf;
 	aufs_bindex_t bindex, bbot;
-	int execed, delayed;
 
-	delayed = (current->flags & PF_KTHREAD) || in_interrupt();
 	finfo = au_fi(file);
 	fidir = finfo->fi_hdir;
 	if (fidir) {
-		au_sphl_del(&finfo->fi_hlist,
-			    &au_sbi(file->f_path.dentry->d_sb)->si_files);
+		au_hbl_del(&finfo->fi_hlist,
+			   &au_sbi(file->f_path.dentry->d_sb)->si_files);
 		vdir_cache = fidir->fd_vdir_cache; /* lock-free */
 		if (vdir_cache)
-			au_vdir_free(vdir_cache, delayed);
+			au_vdir_free(vdir_cache);
 
 		bindex = finfo->fi_btop;
 		if (bindex >= 0) {
-			execed = vfsub_file_execed(file);
 			hf = fidir->fd_hfile + bindex;
 			/*
 			 * calls fput() instead of filp_close(),
@@ -348,12 +346,12 @@ static int aufs_release_dir(struct inode *inode __maybe_unused,
 			bbot = fidir->fd_bbot;
 			for (; bindex <= bbot; bindex++, hf++)
 				if (hf->hf_file)
-					au_hfput(hf, execed);
+					au_hfput(hf, /*execed*/0);
 		}
-		au_delayed_kfree(fidir);
+		kfree(fidir);
 		finfo->fi_hdir = NULL;
 	}
-	au_finfo_fin(file, delayed);
+	au_finfo_fin(file);
 	return 0;
 }
 
@@ -418,7 +416,7 @@ static int au_do_fsync_dir(struct file *file, int datasync)
 	struct super_block *sb;
 	struct inode *inode;
 
-	err = au_reval_and_lock_fdi(file, reopen_dir, /*wlock*/1);
+	err = au_reval_and_lock_fdi(file, reopen_dir, /*wlock*/1, /*fi_lsc*/0);
 	if (unlikely(err))
 		goto out;
 
@@ -479,7 +477,7 @@ static int aufs_iterate_shared(struct file *file, struct dir_context *ctx)
 	struct inode *inode, *h_inode;
 	struct super_block *sb;
 
-	AuDbg("%pD, ctx{%pf, %llu}\n", file, ctx->actor, ctx->pos);
+	AuDbg("%pD, ctx{%ps, %llu}\n", file, ctx->actor, ctx->pos);
 
 	dentry = file->f_path.dentry;
 	inode = d_inode(dentry);
@@ -487,7 +485,7 @@ static int aufs_iterate_shared(struct file *file, struct dir_context *ctx)
 
 	sb = dentry->d_sb;
 	si_read_lock(sb, AuLock_FLUSH);
-	err = au_reval_and_lock_fdi(file, reopen_dir, /*wlock*/1);
+	err = au_reval_and_lock_fdi(file, reopen_dir, /*wlock*/1, /*fi_lsc*/0);
 	if (unlikely(err))
 		goto out;
 	err = au_alive_dir(dentry);
@@ -640,9 +638,9 @@ static int sio_test_empty(struct dentry *dentry, struct test_empty_arg *arg)
 	h_dentry = au_h_dptr(dentry, arg->bindex);
 	h_inode = d_inode(h_dentry);
 	/* todo: i_mode changes anytime? */
-	inode_lock_nested(h_inode, AuLsc_I_CHILD);
+	inode_lock_shared_nested(h_inode, AuLsc_I_CHILD);
 	err = au_test_h_perm_sio(h_inode, MAY_EXEC | MAY_READ);
-	inode_unlock(h_inode);
+	inode_unlock_shared(h_inode);
 	if (!err)
 		err = do_test_empty(dentry, arg);
 	else {
