@@ -1,5 +1,6 @@
+/* SPDX-License-Identifier: GPL-2.0 */
 /*
- * Copyright (C) 2005-2017 Junjiro R. Okajima
+ * Copyright (C) 2005-2018 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -131,11 +132,14 @@ struct au_sbinfo {
 	/* external inode number (bitmap and translation table) */
 	vfs_readf_t		si_xread;
 	vfs_writef_t		si_xwrite;
+	loff_t			si_ximaxent;	/* max entries in a xino */
+
 	struct file		*si_xib;
 	struct mutex		si_xib_mtx; /* protect xib members */
 	unsigned long		*si_xib_buf;
 	unsigned long		si_xib_last_pindex;
 	int			si_xib_next_bit;
+
 	aufs_bindex_t		si_xino_brid;
 	unsigned long		si_xino_jiffy;
 	unsigned long		si_xino_expire;
@@ -144,6 +148,7 @@ struct au_sbinfo {
 
 #ifdef CONFIG_AUFS_EXPORT
 	/* i_generation */
+	/* todo: make xigen file an array to support many inode numbers */
 	struct file		*si_xigen;
 	atomic_t		si_xigen_next;
 #endif
@@ -204,7 +209,7 @@ struct au_sbinfo {
 /*
  * set true when refresh_dirs() failed at remount time.
  * then try refreshing dirs at access time again.
- * if it is false, refreshing dirs at access time is unnecesary
+ * if it is false, refreshing dirs at access time is unnecessary
  */
 #define AuSi_FAILED_REFRESH_DIR	1
 #define AuSi_FHSM		(1 << 1)	/* fhsm is active now */
@@ -345,7 +350,7 @@ int au_test_nfsd(void);
 void au_export_init(struct super_block *sb);
 void au_xigen_inc(struct inode *inode);
 int au_xigen_new(struct inode *inode);
-int au_xigen_set(struct super_block *sb, struct file *base);
+int au_xigen_set(struct super_block *sb, struct path *path);
 void au_xigen_clr(struct super_block *sb);
 
 static inline int au_busy_or_stale(void)
@@ -359,7 +364,7 @@ AuStubInt0(au_test_nfsd, void)
 AuStubVoid(au_export_init, struct super_block *sb)
 AuStubVoid(au_xigen_inc, struct inode *inode)
 AuStubInt0(au_xigen_new, struct inode *inode)
-AuStubInt0(au_xigen_set, struct super_block *sb, struct file *base)
+AuStubInt0(au_xigen_set, struct super_block *sb, struct path *path)
 AuStubVoid(au_xigen_clr, struct super_block *sb)
 AuStub(int, au_busy_or_stale, return -EBUSY, void)
 #endif /* CONFIG_AUFS_EXPORT */
@@ -622,5 +627,11 @@ static inline aufs_bindex_t au_xino_brid(struct super_block *sb)
 	return au_sbi(sb)->si_xino_brid;
 }
 
+static inline loff_t au_xi_maxent(struct super_block *sb)
+{
+	SiMustAnyLock(sb);
+	return au_sbi(sb)->si_ximaxent;
+}
+
 #endif /* __KERNEL__ */
 #endif /* __AUFS_SUPER_H__ */
