@@ -1,5 +1,6 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
- * Copyright (C) 2005-2016 Junjiro R. Okajima
+ * Copyright (C) 2005-2018 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -110,7 +111,7 @@ static void au_nhash_wh_do_free(struct hlist_head *head)
 	struct hlist_node *node;
 
 	hlist_for_each_entry_safe(pos, node, head, wh_hash)
-		au_delayed_kfree(pos);
+		kfree(pos);
 }
 
 static void au_nhash_de_do_free(struct hlist_head *head)
@@ -119,7 +120,7 @@ static void au_nhash_de_do_free(struct hlist_head *head)
 	struct hlist_node *node;
 
 	hlist_for_each_entry_safe(pos, node, head, hash)
-		au_cache_dfree_vdir_dehstr(pos);
+		au_cache_free_vdir_dehstr(pos);
 }
 
 static void au_nhash_do_free(struct au_nhash *nhash,
@@ -137,7 +138,7 @@ static void au_nhash_do_free(struct au_nhash *nhash,
 		nhash_count(head);
 		free(head++);
 	}
-	au_delayed_kfree(nhash->nh_head);
+	kfree(nhash->nh_head);
 }
 
 void au_nhash_wh_free(struct au_nhash *whlist)
@@ -350,23 +351,15 @@ out:
 
 /* ---------------------------------------------------------------------- */
 
-void au_vdir_free(struct au_vdir *vdir, int atonce)
+void au_vdir_free(struct au_vdir *vdir)
 {
 	unsigned char **deblk;
 
 	deblk = vdir->vd_deblk;
-	if (!atonce) {
-		while (vdir->vd_nblk--)
-			au_delayed_kfree(*deblk++);
-		au_delayed_kfree(vdir->vd_deblk);
-		au_cache_dfree_vdir(vdir);
-	} else {
-		/* not delayed */
-		while (vdir->vd_nblk--)
-			kfree(*deblk++);
-		kfree(vdir->vd_deblk);
-		au_cache_free_vdir(vdir);
-	}
+	while (vdir->vd_nblk--)
+		kfree(*deblk++);
+	kfree(vdir->vd_deblk);
+	au_cache_free_vdir(vdir);
 }
 
 static struct au_vdir *alloc_vdir(struct file *file)
@@ -400,10 +393,10 @@ static struct au_vdir *alloc_vdir(struct file *file)
 	if (!err)
 		return vdir; /* success */
 
-	au_delayed_kfree(vdir->vd_deblk);
+	kfree(vdir->vd_deblk);
 
 out_free:
-	au_cache_dfree_vdir(vdir);
+	au_cache_free_vdir(vdir);
 out:
 	vdir = ERR_PTR(err);
 	return vdir;
@@ -415,7 +408,7 @@ static int reinit_vdir(struct au_vdir *vdir)
 	union au_vdir_deblk_p p, deblk_end;
 
 	while (vdir->vd_nblk > 1) {
-		au_delayed_kfree(vdir->vd_deblk[vdir->vd_nblk - 1]);
+		kfree(vdir->vd_deblk[vdir->vd_nblk - 1]);
 		/* vdir->vd_deblk[vdir->vd_nblk - 1] = NULL; */
 		vdir->vd_nblk--;
 	}
@@ -546,7 +539,7 @@ static int au_handle_shwh(struct super_block *sb, struct au_vdir *vdir,
 		}
 	}
 
-	au_delayed_free_page((unsigned long)o);
+	free_page((unsigned long)o);
 
 out:
 	AuTraceErr(err);
@@ -663,7 +656,7 @@ static int read_vdir(struct file *file, int may_read)
 		err = 0;
 		allocated = vdir;
 	} else if (may_read
-		   && (inode->i_version != vdir->vd_version
+		   && (!inode_eq_iversion(inode, vdir->vd_version)
 		       || time_after(jiffies, vdir->vd_jiffy + expire))) {
 		do_read = 1;
 		err = reinit_vdir(vdir);
@@ -679,13 +672,13 @@ static int read_vdir(struct file *file, int may_read)
 	err = au_do_read_vdir(&arg);
 	if (!err) {
 		/* file->f_pos = 0; */ /* todo: ctx->pos? */
-		vdir->vd_version = inode->i_version;
+		vdir->vd_version = inode_query_iversion(inode);
 		vdir->vd_last.ul = 0;
 		vdir->vd_last.p.deblk = vdir->vd_deblk[0];
 		if (allocated)
 			au_set_ivdir(inode, allocated);
 	} else if (allocated)
-		au_vdir_free(allocated, /*atonce*/0);
+		au_vdir_free(allocated);
 
 out:
 	return err;
@@ -776,11 +769,11 @@ int au_vdir_init(struct file *file)
 	inode = file_inode(file);
 	err = copy_vdir(vdir_cache, au_ivdir(inode));
 	if (!err) {
-		file->f_version = inode->i_version;
+		file->f_version = inode_query_iversion(inode);
 		if (allocated)
 			au_set_fvdir_cache(file, allocated);
 	} else if (allocated)
-		au_vdir_free(allocated, /*atonce*/0);
+		au_vdir_free(allocated);
 
 out:
 	return err;
@@ -849,7 +842,8 @@ static int seek_vdir(struct file *file, struct dir_context *ctx)
 
 out:
 	/* smp_mb(); */
-	AuTraceErr(!valid);
+	if (!valid)
+		AuDbg("valid %d\n", !valid);
 	return valid;
 }
 
